#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Colony Framework
# Copyright (c) 2008-2020 Hive Solutions Lda.
#
# This file is part of Hive Colony Framework
#
# Hive Colony Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Colony Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Colony Framework If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2020 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

def visit(ast_node_class):

    def decorator(function, *args, **kwargs):
        function.ast_node_class = ast_node_class
        return function

    return decorator

def dispatch_visit(map_name = "node_method_map"):

    def create_interceptor(function):

        def decorator_interceptor(*args, **kwargs):
            # unpacks the first two "unnamed" arguments as the self
            # instance reference and the node element to be visited
            self = args[0]
            node = args[1]

            # verifies if the current instance contains the node method
            # map if that's the case retrieves otherwise falls back to an
            # empty dictionary (for code compatibility)
            has_map = hasattr(self, map_name)
            node_method_map = self.node_method_map if has_map else dict()

            # retrieves the class for the node argument and then
            # gathers the complete mro class definition to be able
            # to iterate over the class hierarchy
            node_class = node.__class__
            mro = node_class.mro()

            # iterates over the complete class hierarchy for the provided
            # node (from bottom to up) so that the best match for the
            # visit operation is found and properly called
            for mro_item in mro:
                # in case the current mro item class level is nor found
                # skips the current iteration (cannot visit at this level)
                if not mro_item in node_method_map: continue

                # the current class level is valid and so the proper method
                # is retrieved from the map and then called with the provided
                # arguments, note that a before visit and an after visit calls
                # are done so that proper "notification" exists
                visit_method = node_method_map[mro_item]
                self.before_visit(*args[1:], **kwargs)
                visit_method(*args, **kwargs)
                self.after_visit(*args[1:], **kwargs)

                # returns the control flow to the caller method, note that no
                # value is returned as this is a simple visit operation
                return

            # in case no visit has been made a normal fallback operation is
            # performed by calling the "original" function/method
            return function(*args, **kwargs)

        return decorator_interceptor

    def decorator(function, *args, **kwargs):
        interceptor = create_interceptor(function)
        return interceptor

    return decorator



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Colony Framework
# Copyright (c) 2008-2020 Hive Solutions Lda.
#
# This file is part of Hive Colony Framework
#
# Hive Colony Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Colony Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Colony Framework If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt> & Tiago Silva <tsilva@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2020 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import os
import sys
import glob
import getopt
import warnings

import colony

USAGE = "Help:\n\
--help[-h] - prints this message\n\
--noloop[-n] - sets the manager to not use the loop mode\n\
--level[-v]=(LEVEL) - sets the logging verbosity level to be used\n\
--layout_mode[-l]=development/repository_svn/production - sets the layout mode to be used\n\
--run_mode[-r]=development/test/production - sets the run mode to be used\n\
--container[-c]=default - sets the container to be used\n\
--daemon_pid[-o]=(DAEMON_PID) - sets the pid of the parent daemon\n\
--config_file[-f]=(CONFIGURATION_FILE) - sets the file path to the configuration file\n\
--daemon_file[-d]=(DAEMON_FILE) - sets the file path to the daemon file\n\
--manager_dir[-m]=(PLUGIN_DIR) - sets the plugin directory to be used by the manager\n\
--logger_dir[-g]=(LOGGER_DIR) - sets the logger directory to be used by the manager for the logger\n\
--library_dir[-i]=(LIBRARY_DIR_1;LIBRARY_DIR_2;...) - sets the series of library directories to use\n\
--plugin_dir[-p]=(PLUGIN_DIR_1;PLUGIN_DIR_2;...) - sets the series of plugin directories to use"
""" The usage string for the command line arguments,
this is going to be display as part of the help string """

BRANDING_TEXT = "Hive Colony %s (Hive Solutions Lda. r%s:%s %s)"
""" The branding text value to be display at the start
of the process execution (contains user/system information) """

VERSION_PRE_TEXT = "Python "
""" The version pre text value """

HELP_TEXT = "Type \"help\" for more information."
""" The help text value """

DEFAULT_CONFIG_FILE_PATH = "config/python/devel.py"
""" The default configuration file path, that is going
to be used only in case it is available for under the
provided path, otherwise only the default base values
are used (security setting) """

RELATIVE_MANAGER_PATH = os.path.dirname(os.path.realpath(__file__))
""" The default manager path, considered to be the current
executing file's directory (by default)) """

DEFAULT_LOGGER_PATH = "log"
""" The default logger path """

PLUGIN_PATHS_FILE = "plugins.pth"
""" The colony plugin paths file """

# registers the ignore flag in the deprecation warnings so that
# no message with this kind of warning is printed (clean console)
warnings.filterwarnings("ignore", category = DeprecationWarning)

def usage():
    """
    Prints the usage for the command line, should be used
    to provide extra information about the way to use this
    command line to the user.
    """

    print(USAGE)

def print_information():
    """
    Prints the system information for the command line.
    This should be the first string presented to the end
    used when running the system.
    """

    print(BRANDING_TEXT % (
        colony.VERSION,
        colony.RELEASE,
        colony.BUILD,
        colony.RELEASE_DATE
    ))
    print(VERSION_PRE_TEXT + sys.version)
    print(HELP_TEXT)

def run(
    manager_path,
    logger_path,
    library_path,
    meta_path,
    plugin_path,
    mode = None,
    margs = [],
    level = "WARNING",
    layout_mode = "default",
    run_mode = "default",
    stop_on_cycle_error = True,
    loop = False,
    threads = True,
    signals = True,
    container = "default",
    prefix_paths = [],
    daemon_pid = None,
    daemon_file_path = None
):
    """
    Starts the loading of the plugin manager. This should be the
    primary start point of the plugin system when starting it as
    a stand alone process (eg: not using wsgi).

    :type manager_path: String
    :param manager_path: The manager base path for execution.
    :type logger_path: String
    :param logger_path: The manager base path for logger.
    :type library_path: String
    :param library_path: The set of paths to the various library
    locations separated by a semi-column.
    :type meta_path: String
    :param meta_path: The set of paths to the various meta
    locations separated by a semi-column.
    :type plugin_path: String
    :param plugin_path: The set of paths to the various plugin
    locations separated by a semi-column.
    :type mode: String
    :param mode: The mode that is going to be used for non
    standard execution of the plugin manager (eg: testing). This
    value is not set by default (for default execution).
    :type margs: List
    :param margs: The arguments (coming from command line) that
    are going to be provided at execution time to the function
    responsible for the mode execution, these arguments should
    consist on a series of string based values.
    :type level: String
    :param level: The logging level described as a string that is
    going to be used by the underlying colony logging infra-structure.
    :type layout_mode: String
    :param layout_mode: The layout mode to be used by the plugin system.
    :type run_mode: String
    :param run_mode: The run mode to be used by the plugin
    system, this value is critical for the type of execution
    of the colony system (eg: development, runtime, etc.)
    :type stop_on_cycle_error: bool
    :param stop_on_cycle_error: If the plugin system should stop
    on cycle error, a cycle error is an error that occurs during
    the startup process of the colony infra-structure.
    :type loop: bool
    :param loop: If the plugin manager is going to run in a loop.
    :type threads: bool
    :param threads: If the plugin manager is going to allow threads.
    :type signals: bool
    :param signals: If the plugin manager is going to register signals.
    :type container: String
    :param container: The name of the plugin manager container.
    :type prefix_paths: List
    :param prefix_paths: The list of manager path relative paths to be used as reference for sub-projects.
    :type daemon_pid: int
    :param daemon_pid: The pid of the daemon process running the instance of plugin manager.
    :type daemon_file_path: String
    :param daemon_file_path: The file path to the daemon file, for information control.
    :rtype: int
    :return: The return code.
    """

    # sets the plugin manager as a global variable, this value
    # may be used in more situations that the one defined in the
    # the current (local) scope
    global plugin_manager

    # checks if the library path is not valid, by splitting its
    # value around the separator token into a valid list
    if library_path: library_paths = library_path.split(";")
    else: library_paths = []

    # checks if the meta path is not valid, by splitting its
    # value around the separator token into a valid list
    if meta_path: meta_paths = meta_path.split(";")
    else: meta_paths = []

    # checks if the plugin path is not valid, by splitting its
    # value around the separator token into a valid list
    if plugin_path: plugin_paths = plugin_path.split(";")
    else: plugin_paths = []

    # retrieves the current executing platform, so that this
    # value may be passed to the next functions to be called
    # as part of this execution stack
    platform = colony.get_environment()

    # runs the ensure operation for the currently defined manager
    # path making sure that the complete directory structure exists
    colony.ensure_tree(manager_path)

    # creates the plugin manager with the given plugin paths, this
    # is the instance that is going to be used for the current loading
    plugin_manager = colony.PluginManager(
        manager_path = manager_path,
        logger_path = logger_path,
        library_paths = library_paths,
        meta_paths = meta_paths,
        plugin_paths = plugin_paths,
        platform = platform,
        init_complete_handlers = [],
        stop_on_cycle_error = stop_on_cycle_error,
        loop = loop,
        threads = threads,
        signals = signals,
        layout_mode = layout_mode,
        run_mode = run_mode,
        container = container,
        prefix_paths = prefix_paths,
        daemon_pid = daemon_pid,
        daemon_file_path = daemon_file_path
    )

    # resolves the string based level into the proper integer
    # that describes the logging level and then uses that value
    # to start the logging infra-structure of colony
    level = colony.getLevelName(level)
    plugin_manager.start_logger(level)

    # creates the callback function to be used in the process of
    # printing the branding information text to the standard output
    # informing the end user about the current environment
    callback = lambda: print_information()

    # starts and loads the plugin system, this is a blocking
    # call and the flow control is only returned at the end of
    # the execution of the colony infra-structure, then the
    # returned code is returned to the caller function
    return_code = plugin_manager.load_system(
        mode = mode,
        args = margs,
        callback = callback
    )
    return return_code

def execute(cwd = None, force_exit = True):
    """
    The main entry point of the application, should parse
    the provided command line arguments and then start the
    execution of the colony plugin system.

    An optional force exit flag controls if the exit function
    should always be used in exit.

    :type cwd: String
    :param cwd: The "original" current working directory to
    be used for situations where the "cwd" has been changed
    so that files generated are put on the colony path. This
    is created as a legacy operation.
    :type force_exit: bool
    :param force_exit: If in case the return code is a valid
    one, the exit function should "still" be used to return
    the control flow immediately to the caller process.
    """

    # verifies if the cwd value is defined an in case it's not
    # retrieves the real cwd values as to be used for operations
    cwd = cwd or os.getcwd()

    try:
        # defines the options retrieval schema/template and runs it
        # retrieves the various options and the remaining arguments
        # that have not been parsed by the processor
        options, args = getopt.getopt(
            sys.argv[1:],
            "hnv:l:r:c:o:f:d:m:g:i:t:p:",
            [
                 "help"
                 "noloop",
                 "level="
                 "layout_mode=",
                 "run_mode=",
                 "container=",
                 "daemon_pid=",
                 "config_file=",
                 "daemon_file=",
                 "manager_dir=",
                 "logger_dir=",
                 "library_dir=",
                 "meta_dir=",
                 "plugin_dir="
            ]
        )
    except getopt.GetoptError as error:
        # prints the error description so that the user is able
        # to react to the error, then prints the possible usage
        # for the command and exists in error
        print(str(error))
        usage()
        sys.exit(2)

    # retrieves the execution mode for colony as the first non parsed
    # value from the command line (as expected)
    if args: mode = args[0]
    else: mode = None

    # retrieves the complete set of arguments that are going to be
    # provided for the mode execution, these arguments are considered
    # to be (mode) context specific and valuable only inside context
    margs = args[1:]

    # retrieves the file system encoding
    file_system_encoding = sys.getfilesystemencoding()

    # starts the options values
    loop = True
    level = None
    threads = True
    signals = True
    layout_mode = None
    run_mode = None
    container = "default"
    daemon_pid = None
    config_file_path = DEFAULT_CONFIG_FILE_PATH
    daemon_file_path = None
    manager_path = colony.resolve_manager(RELATIVE_MANAGER_PATH)
    logger_path = DEFAULT_LOGGER_PATH
    library_path = None
    meta_path = None
    plugin_path = None

    # iterates over all the options to be able to parse its value
    # starting it from the command line
    for option, value in options:
        if option in ("-h", "--help"):
            usage()
            sys.exit()
        elif option in ("-n", "--noloop"):
            loop = False
        elif option in ("-v", "--level"):
            level = value
        elif option in ("-l", "--layout_mode"):
            layout_mode = value
        elif option in ("-r", "--run_mode"):
            run_mode = value
        elif option in ("-c", "--container"):
            container = value
        elif option in ("-o", "--daemon_pid"):
            daemon_pid = int(value)
        elif option in ("-f", "--config_file"):
            config_file_path = value.decode(file_system_encoding)
        elif option in ("-d", "--daemon_file"):
            daemon_file_path = value.decode(file_system_encoding)
        elif option in ("-m", "--manager_dir"):
            manager_path = value.decode(file_system_encoding)
        elif option in ("-g", "--logger_dir"):
            logger_path = value.decode(file_system_encoding)
        elif option in ("-i", "--library_dir"):
            library_path = value.decode(file_system_encoding)
        elif option in ("-t", "--meta_dir"):
            meta_path = value.decode(file_system_encoding)
        elif option in ("-p", "--plugin_dir"):
            plugin_path = value.decode(file_system_encoding)
        else:
            assert False, "unhandled option"

    # parses the configuration options, retrieving the various values that
    # control the execution of the plugin system
    mode, level, layout_mode, run_mode, stop_on_cycle_error,\
    prefix_paths, daemon_file_path, logger_path, library_path, meta_path,\
    plugin_path = parse_configuration(
        cwd,
        mode,
        config_file_path,
        level,
        layout_mode,
        run_mode,
        daemon_file_path,
        logger_path,
        library_path,
        meta_path,
        plugin_path,
        manager_path
    )

    # configures the system using the layout mode, the run mode
    # and the  manager path
    configure_system(layout_mode, run_mode, manager_path)

    # in case the daemon file path is valid and not an absolute path
    # must the (complete) daemon file path prepending the manager path
    if daemon_file_path and not os.path.isabs(daemon_file_path):
        daemon_file_path = manager_path + "/" + daemon_file_path

    # in case the logger path is not an absolute path, must create
    # the (complete) logger path prepending the manager path
    if not os.path.isabs(logger_path): logger_path = manager_path + "/" + logger_path

    # strips the various component location paths around the
    # semi-colon character so that it's possible to send them
    library_path_striped = library_path.strip(";")
    meta_path_striped = meta_path.strip(";")
    plugin_path_striped = plugin_path.strip(";")

    # starts the running process, this should launch the manager
    # and then start the main loop of execution returning the
    # return code (result of execution) to the caller process
    return_code = run(
        manager_path,
        logger_path,
        library_path_striped,
        meta_path_striped,
        plugin_path_striped,
        mode = mode,
        margs = margs,
        level = level,
        layout_mode = layout_mode,
        run_mode = run_mode,
        stop_on_cycle_error = stop_on_cycle_error,
        loop = loop,
        threads = threads,
        signals = signals,
        container = container,
        prefix_paths = prefix_paths,
        daemon_pid = daemon_pid,
        daemon_file_path = daemon_file_path
    )

    # in case the return code is not success or the force
    # exit flag is set then calls the exit function
    if not return_code == 0 or force_exit: exit(return_code)

def parse_configuration(
    cwd,
    mode,
    config_file_path,
    level,
    layout_mode,
    run_mode,
    daemon_file_path,
    logger_path,
    library_path,
    meta_path,
    plugin_path,
    manager_path
):
    """
    Parses the configuration using the given values as default values.
    The configuration file used is given as a parameter to the function.

    :type cwd: String
    :param cwd: The (original) current working directory to be used
    in the resolution of the relative import values.
    :type mode: String
    :param mode: The mode that is going to be used for non
    standard execution of the plugin manager (eg: testing). This
    value is not set by default (for default execution).
    :type config_file_path: Sting
    :param config_file_path: The path to the configuration file.
    :type level: String
    :param level: The logging level value described as a string
    that is going to be used in the plugin system.
    :type layout_mode: String
    :param layout_mode: The layout mode to be used by the plugin system.
    :type run_mode: String
    :param run_mode: The run mode to be used by the plugin system.
    :type daemon_file_path: String
    :param daemon_file_path: The file path to the daemon file,
    for information control.
    :type logger_path: String
    :param logger_path: The path to the logger.
    :type library_path: String
    :param library_path: The set of paths to the various library
    locations separated by a semi-column.
    :type meta_path: String
    :param meta_path: The set of paths to the various meta locations
    separated by a semi-column.
    :type plugin_path: String
    :param plugin_path: The set of paths to the various plugin
    locations separated by a semi-column.
    :type manager_path: String
    :param manager_path: The path to the plugin system.
    :rtype: Tuple
    :return: The tuple with the values parsed value.
    """

    # retrieves the configuration directory from the configuration
    # file path (the directory is going to be used to include the module)
    config_dir = os.path.dirname(config_file_path)

    # in case the configuration directory path is not an absolute path
    if not os.path.isabs(config_dir):
        # creates the (complete) configuration directory path
        # prepending the manager path
        config_dir = os.path.normpath(manager_path + "/" + config_dir)

    # in case the configuration directory path is valid inserts it into the system path
    config_dir and sys.path.insert(0, config_dir)

    # retrieves the configuration file base path from the configuration file path
    config_file_base_path = os.path.basename(config_file_path)

    # retrieves the configuration module name and the configuration
    # module extension by splitting the configuration base path into
    # base name and extension and then imports the referring module
    config_module, _configuration_module_extension = os.path.splitext(config_file_base_path)
    try: config = __import__(config_module)
    except ImportError: import colony.config.base as module; config = module

    # retrieves the contents of the configuration file that has just
    # been loaded, this is the default operation to be performed
    names = dir(config)

    # sets the proper configuration attributes taking into account the
    # presence or not of such attributes in the loaded file
    if "level" in names: level = level or config.level
    if "layout_mode" in names: layout_mode = layout_mode or config.layout_mode
    if "run_mode" in names: run_mode = run_mode or config.run_mode
    if "prefix_paths" in names: prefix_paths = config.prefix_paths
    if "stop_on_cycle_error" in names: stop_on_cycle_error = config.stop_on_cycle_error
    if "daemon_file_path" in names: daemon_file_path = config.daemon_file_path
    if "logger_path" in names: logger_path = config.logger_path

    # tries to load some of the environment "loadable" properties for the
    # starting of the colony infra-structure, defaulting to the provided
    # values in case they are present (as expected), then in case there's
    # still no valid values for such variables default values are used
    mode = colony.conf("MODE", mode)
    level = colony.conf("LEVEL", level)
    layout_mode = colony.conf("LAYOUT_MODE", layout_mode)
    run_mode = colony.conf("RUN_MODE", run_mode)
    layout_mode = layout_mode or "default"
    run_mode = run_mode or "default"

    # retrieves the complete set of configuration variables associated
    # with the various paths to be used by colony and then adds the
    # proper static file based configuration values to them, so that
    # these list are properly started with the initial values
    library_path_list = colony.conf("LIBRARY_PATH", [], cast = list)
    meta_path_list = colony.conf("META_PATH", [], cast = list)
    plugin_path_list = colony.conf("PLUGIN_PATH", [], cast = list)
    library_path_list = library_path_list + config.library_path_list
    meta_path_list = meta_path_list + config.meta_path_list
    plugin_path_list = plugin_path_list + config.plugin_path_list

    # in case the library path is defined, must appends a
    # separator to the library path to mark the initial separation
    # otherwise creates a new library path string initializing the
    # value to an empty string so that it can be extended
    if library_path: library_path += ";"
    else: library_path = ""

    # in case the meta path is defined, must appends a
    # separator to the meta path to mark the initial separation
    # otherwise creates a new meta path string initializing the
    # value to an empty string so that it can be extended
    if meta_path: meta_path += ";"
    else: meta_path = ""

    # in case the plugin path is defined, must appends a
    # separator to the plugin path to mark the initial separation
    # otherwise creates a new plugin path string initializing the
    # value to an empty string so that it can be extended
    if plugin_path: plugin_path += ";"
    else: plugin_path = ""

    # retrieves the current prefix paths
    current_prefix_paths = prefix_paths[layout_mode]

    # retrieves the extra library path as the dereferenced values
    # from the colony configuration library path list and adds the
    # extra library path to the library path
    extra_library_path = convert_reference_path_list(
        cwd,
        manager_path,
        current_prefix_paths,
        library_path_list
    )
    library_path += extra_library_path

    # retrieves the extra meta path as the dereferenced values
    # from the colony configuration meta path list and adds the
    # extra meta path to the meta path
    extra_meta_path = convert_reference_path_list(
        cwd,
        manager_path,
        current_prefix_paths,
        meta_path_list
    )
    meta_path += extra_meta_path

    # loads the plugin paths file path and adds the plugin paths
    # file path to the plugin path
    plugin_paths_file_path = load_plugin_paths_file(manager_path)
    plugin_path += plugin_paths_file_path

    # retrieves the extra plugin path as the dereferenced values
    # from the colony configuration plugin path list and adds the
    # extra plugin path to the plugin path
    extra_plugin_path = convert_reference_path_list(
        cwd,
        manager_path,
        current_prefix_paths,
        plugin_path_list
    )
    plugin_path += extra_plugin_path

    return (
        mode,
        level,
        layout_mode,
        run_mode,
        stop_on_cycle_error,
        current_prefix_paths,
        daemon_file_path,
        logger_path,
        library_path,
        meta_path,
        plugin_path
    )

def convert_reference_path_list(cwd, manager_path, current_prefix_paths, reference_path_list):
    """
    Converts the given list of reference paths. The reference
    paths include references of type %reference_name% to include
    base paths that are dereferenced at runtime based in the current
    layout configuration or other variables.

    :type cwd: String
    :param cwd: The (original) current working directory to be used
    in the resolution of the relative import values.
    :type manager_path: String
    :param manager_path: The path to the manager.
    :type current_prefix_paths: List
    :param current_prefix_paths: The prefix paths currently in use.
    :type reference_path_list: List
    :param reference_path_list: The list of reference paths.
    :rtype: String
    :return: A string converted reference path containing all
    the dereferenced paths.
    """

    # initializes the converted reference path, this is the value
    # to be returned by this function with the complete path string
    # to be used by the plugin system
    converted_reference_path = str()

    # iterates over all the reference paths, in order to normalize
    # resolver and integrate them into the reference path
    for reference_path in reference_path_list:
        # verifies if the current reference path is a local one (using
        # the prefix value) and in case it's not prepends the manager
        # path to it as the base path (usual situation)
        is_local = reference_path.startswith("./")
        if is_local: dereferenced_path = cwd + "/" + reference_path
        else: dereferenced_path = manager_path + "/" + reference_path

        # iterates over all the current prefix paths to dereference
        # them into the path "along" the various prefix paths
        for current_prefix_path in current_prefix_paths:
            # retrieves the current prefix path name and value
            # to be used in the dereferencing of the path, then
            # executes the dereferencing operation substituting the
            # "wildcard" references in the paths
            current_prefix_path_name = "%" + current_prefix_path + "_prefix_path%"
            current_prefix_path_value = current_prefix_paths[current_prefix_path]
            dereferenced_path = dereferenced_path.replace(
                current_prefix_path_name,
                current_prefix_path_value
            )

        # runs the glob based resolver to resolver the "wildcard" patterns
        # that may be present in the path, this operation should return
        # a list of paths from the resolved "wildcard" them iterates over
        # these paths to add them to the converted reference path
        dereferenced_paths = glob.glob(dereferenced_path)
        for dereferenced_path in dereferenced_paths:
            # resolves the dereferenced path as an absolute path and
            # adds it to the converted reference string path (linear
            # version of the path separated by tokens)
            dereferenced_path = os.path.abspath(dereferenced_path)
            converted_reference_path += dereferenced_path + ";"

    # returns the converted reference path
    return converted_reference_path

def load_plugin_paths_file(manager_path):
    """
    Loads the plugin paths file, creating the base plugin
    paths contained in the file.

    :type manager_path: String
    :param manager_path: The path to the manager.
    :rtype: String
    :return: A string with the paths loaded from the file.
    """

    # creates the config general path from the manager path
    config_general_path = manager_path + "/config/general"

    # crates the plugin paths file path (from the config general path)
    plugin_paths_file_path = config_general_path + "/" + PLUGIN_PATHS_FILE

    # in case the plugin paths file does not exists (the
    # file is not mandatory) must return immediately because
    # no further processing is taking place
    if not os.path.exists(plugin_paths_file_path): return ""

    # opens the plugin paths file for reading then reads the
    # plugin path files contents, and the closes file to avoid
    # any further reading that could cause memory leaks
    plugin_paths_file = open(plugin_paths_file_path, "r")
    try: plugin_paths_file_contents = plugin_paths_file.read()
    finally: plugin_paths_file.close()

    # splits the paths over the newline character and then
    # filters the invalid values (white spaces)
    paths = plugin_paths_file_contents.split("\n")
    paths = [value for value in paths if value]

    # initializes the converted reference path
    plugin_paths_string_value = str()

    # iterates over all the paths to creates the
    # plugins paths string
    for path in paths:
        # in case the path is not an absolute path
        # creates the (complete) path prepending
        # the manager path and then adds the path
        # to the plugin paths string value
        if not os.path.isabs(path): path = manager_path + "/" + path
        plugin_paths_string_value += path + ";"

    # returns the plugin paths string value
    return plugin_paths_string_value

def configure_system(layout_mode, run_mode, manager_path):
    """
    Configures the system for the given attributes.

    :type layout_mode: String
    :param layout_mode: The layout mode to configure the system.
    :type run_mode: String
    :param run_mode: The run mode to configure the system.
    :type manager_path: String
    :param manager_path: The manager path to configure the system.
    """

    # sets the various colony related environment variables
    # so that the current process may expose them to any
    # created child process (context exposure)
    os.environ["COLONY_LAYOUT_MODE"] = layout_mode
    os.environ["COLONY_RUN_MODE"] = run_mode
    os.environ["COLONY_HOME"] = manager_path

    # constructs the library path and normalizes it
    library_path = manager_path + "/colony/libs"
    library_path = os.path.normpath(library_path)

    # inserts the library path into the system path
    sys.path.insert(0, library_path)

def main():
    """
    Execution function for the colony infra-structure may be used
    for the entry point definition.

    Note that a call to this function may change the current working
    directory for the executing process.
    """

    path = os.path.dirname(__file__)
    path = os.path.normpath((os.path.abspath(path)))
    _path = os.getcwd()
    try:
        os.chdir(path)
        execute(cwd = _path)
    finally:
        os.chdir(_path)

if __name__ == "__main__":
    main()
else:
    __path__ = []



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Colony Framework
# Copyright (c) 2008-2020 Hive Solutions Lda.
#
# This file is part of Hive Colony Framework
#
# Hive Colony Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Colony Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Colony Framework If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2020 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import math
import decimal

def get_number_length(number):
    """
    Retrieves the length (in digits) of the given number.
    This method does not allow any kind of padding zeros
    as they are not going to be handled.

    :type number: int
    :param number: The number to retrieve the length
    in number of digits.
    :rtype: int
    :return: The length of the given number in number
    of digits, according to logarithmic calculus.
    """

    return int(math.ceil(math.log(number, 10)))

def get_digit(number, index):
    """
    Retrieves the digit in the given index.
    The index values start with the least
    significant value.

    :type number: int
    :param number: The base number to retrieve
    the digit.
    :type index: int
    :param index: The index to retrieve the digit.
    :rtype: int
    :return: The digit in the given index.
    """

    return number // (10 ** index) % 10

def to_fixed(number, places = 2):
    """
    Converts a (possible) float number into a decimal
    representation of fixed point real.

    This function is useful for situation where operations
    on the float number rendered it impossible for a
    precise comparison.

    Use this method carefully as it may provide unrealistic
    comparisons.

    :type number: float
    :param number: The float number to be converted in to
    a fixed point decimal number.
    :type places: int
    :param places: The number of decimal places to be used
    in the conversion into fixed point.
    :rtype: Decimal
    :return: The decimal representation of the number (fixed
    point arithmetic) with the correct number of decimal places.
    """

    # creates the "initial" format string with the correct
    # (requested) number of decimal places
    format_string = "%%.%df" % places

    # uses the created format string to create a string
    # representing the number with the request decimal places
    # and the uses it as the argument to construct the correct
    # decimal representation of the number
    number_string = format_string % number
    number_fixed = decimal.Decimal(number_string)

    # returns the created decimal representation of the number
    # in a fixed point arithmetic approach
    return number_fixed



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Colony Framework
# Copyright (c) 2008-2020 Hive Solutions Lda.
#
# This file is part of Hive Colony Framework
#
# Hive Colony Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Colony Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Colony Framework If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2020 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import os
import signal
import subprocess

WINDOWS_KILL_COMMAND = "taskkill /F /T /PID %i"
""" The windows kill command """

NT_PLATFORM_VALUE = "nt"
""" The nt platform value """

DOS_PLATFORM_VALUE = "dos"
""" The dos platform value """

WINDOWS_PLATFORMS_VALUE = (
    NT_PLATFORM_VALUE,
    DOS_PLATFORM_VALUE
)
""" The windows platform value """

def kill_process(pid, signal = None):
    """
    Kills the process with the given pid (process identifier)
    and "using" the given signal.

    :type pid: int
    :param pid: The identifier of the process to be killed.
    :type signal: signal
    :param signal: The signal to be used to kill the process.
    """

    # retrieves the current os name
    os_name = os.name

    # in case the current operative system is windows based
    if os_name in WINDOWS_PLATFORMS_VALUE:
        # kills the process using windows methods
        _kill_process_windows(pid)
    # otherwise it must be a unix platform
    else:
        # kills the process using unix methods
        _kill_process_unix(pid)

def _kill_process_windows(pid):
    """
    Kills the process with the given pid (process identifier).
    This class focus in the strategy used in windows operative
    systems.

    :type pid: int
    :param pid: The identifier of the process to be killed.
    """

    # in case the os module contains the kill method
    if hasattr(os, "kill"):
        # kills the process with the given pid
        os.kill(pid, signal.SIGTERM)
    else:
        # crates a process to kill the process with the given pid
        subprocess.Popen(WINDOWS_KILL_COMMAND % pid, shell = True)

def _kill_process_unix(pid, _signal = None):
    """
    Kills the process with the given pid (process identifier).
    This class focus in the strategy used in unix platforms.

    :type pid: int
    :param pid: The identifier of the process to be killed.
    """

    # sets the signal value
    _signal = _signal or signal.SIGKILL #@UndefinedVariable

    # kills the process with the given pid
    # and with the given signal
    os.kill(pid, _signal) #@UndefinedVariable



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Colony Framework
# Copyright (c) 2008-2020 Hive Solutions Lda.
#
# This file is part of Hive Colony Framework
#
# Hive Colony Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Colony Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Colony Framework If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2020 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import os

from colony.base import legacy

try: import Crypto.Cipher.AES
except ImportError: Crypto = None

BLOCK_SIZE = 16
""" The block size to be used for the post operation
should not be too small or security issues may arise """

class AesCipher(object):
    """
    The class responsible for a proper encryption
    and decryption system for the aes system.

    The aes include the padding infra-structure for
    according to pkcs5.

    :see: http://tools.ietf.org/html/rfc2898
    """

    key = None
    """ The symmetric key to be use to encrypt
    messages under this aes cipher """

    block_size = BLOCK_SIZE
    """ The size of the encryption block to be
    used under this cipher instance """

    def __init__(self, key = None, block_size = BLOCK_SIZE):
        """
        Constructor of the class.

        :type key: String
        :param key: The symmetric key (secret) to be used
        in the aes encryption, in case it's not defined
        a new random key will be created.
        :type block_size: int
        :param block_size: The size of the encryption
        block to be used under this cipher instance, must
        be a multiple of eight.
        """

        self.key = key or os.urandom(block_size)
        self.block_size = block_size

    def encrypt(self, raw):
        """
        Encrypts the provided raw string value according
        to the aes and pkcs5 specifications.

        :type raw: String
        :param raw: The raw string value to be used in
        the encryption process.
        :rtype: String
        :return: The encrypted string according to the aes
        cryptographic system in ecb mode.
        """

        raw = self.pad(raw)
        cipher = Crypto.Cipher.AES.new(self.key, Crypto.Cipher.AES.MODE_ECB)
        return cipher.encrypt(raw)

    def decrypt(self, encoded):
        """
        Decrypts the provided encoded (encrypted) string
        into the raw value.

        :type encoded: String
        :param encoded: The encrypted string to be used
        in the decryption process, should be padded according
        to the pkcs5 schema.
        :rtype String
        :return: The decrypted string according to the aes
        cryptographic system in ecb mode.
        """

        cipher = Crypto.Cipher.AES.new(self.key, Crypto.Cipher.AES.MODE_ECB)
        decoded = cipher.decrypt(encoded)
        return self.unpad(decoded)

    def pad(self, value):
        """
        Adds the pkcs5 padding to the provided value
        it should add all the extra padding values to it.

        :type value: String
        :param value: The base value for which the padded
        characters will be added.
        :rtype: String
        :return: The pkcs5 padded string with the padding
        characters added to it.
        """

        remaining = self.block_size - len(value) % self.block_size
        padding = remaining * legacy.chr(remaining)
        return value + padding

    def unpad(self, value):
        """
        Removes the pkcs5 padding from the provided value
        it should remove all the extra padding values from it.

        :type value: String
        :param value: The padded value from which all the
        extra padding characters.
        :rtype: String
        :return: The sanitized string without the extra
        padding characters.
        """

        last = value[-1]
        pad_size = ord(last)
        return value[:-pad_size]

    def get_key(self):
        """
        Retrieves the symmetric key that is currently set
        in the aes instance and that is going to be used
        for encryption and decryption, it should be of the
        same size as the block.

        :rtype: String
        :return: The symmetric key that is currently set
        in the aes instance.
        """

        return self.key

    def get_block_size(self):
        """
        Retrieves the size of the encryption block to be
        used for encryption and decryption, this value is
        used for both the generation of the key and for
        the padding creation.

        The integer value should be a multiple of eight.

        :rtype: int
        :return: The size of the encryption block to be
        used for encryption and decryption.
        """

        return self.block_size



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Colony Framework
# Copyright (c) 2008-2020 Hive Solutions Lda.
#
# This file is part of Hive Colony Framework
#
# Hive Colony Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Colony Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Colony Framework If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2020 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import colony

class RoundTest(colony.ColonyTestCase):
    """
    Test case bundle aimed at verifying the integrity of the
    round associated functions.
    """

    def test_legacy(self):
        is_new = colony.round_is_new()

        result = round(2.675, 2)
        self.assertEqual(result, 2.67 if is_new else 2.68)

        result = round(2.685, 2)
        self.assertEqual(result, 2.69)

    def test_roundi(self):
        is_new = colony.round_is_new()

        result = colony.roundi(2.675, 2)
        self.assertEqual(result, 2.68)

        result = colony.roundi(2.685, 2)
        self.assertEqual(result, 2.69)

        result = colony.roundi(2.68, 2)
        self.assertEqual(result, 2.68)

        result = colony.roundi(2.683, 2)
        self.assertEqual(result, 2.68)

        result = colony.roundi(2.689, 2)
        self.assertEqual(result, 2.69)

        result = colony.roundi(2.695, 2)
        self.assertEqual(result, 2.70)

        result = colony.roundi(2.999, 2)
        self.assertEqual(result, 3.0)

        result = colony.roundi(2.945, 2)
        self.assertEqual(result, 2.95)

        result = colony.roundi(2.9444, 2)
        self.assertEqual(result, 2.94)

        result = colony.roundi(2.9444444444444444444444, 2)
        self.assertEqual(result, 2.94)

        result = colony.roundi(2.9944444444444444444444, 2)
        self.assertEqual(result, 2.99)

        result = colony.roundi(2.995, 2)
        self.assertEqual(result, 3.0)

        result = colony.roundi(1171.735, 2)
        self.assertNotEqual(result, 1171.74)

        result = colony.roundi(99999999.995, 2)
        self.assertEqual(result, 100000000.0)

        result = colony.roundi(999999999999999999999999999.9944444444444444444444, 2)
        if is_new: self.assertEqual(result, 999999999999999999999999999.99)

        result = colony.roundi(999999999999999999999999999.995, 2)
        if is_new: self.assertEqual(result, 1000000000000000000000000000.0)

        result = colony.roundi(770.155, 2)
        if is_new: self.assertEqual(result, 770.15)
        else: self.assertEqual(result, 770.16)

    def test_rounds(self):
        is_new = colony.round_is_new()

        result = colony.rounds(2.675, 2)
        self.assertEqual(result, 2.68)

        result = colony.rounds(2.685, 2)
        self.assertEqual(result, 2.69)

        result = colony.rounds(2.68, 2)
        self.assertEqual(result, 2.68)

        result = colony.rounds(2.683, 2)
        self.assertEqual(result, 2.68)

        result = colony.rounds(2.689, 2)
        self.assertEqual(result, 2.69)

        result = colony.rounds(2.695, 2)
        self.assertEqual(result, 2.70)

        result = colony.rounds(2.999, 2)
        self.assertEqual(result, 3.0)

        result = colony.rounds(2.945, 2)
        self.assertEqual(result, 2.95)

        result = colony.rounds(2.9444, 2)
        self.assertEqual(result, 2.94)

        result = colony.rounds(2.9444444444444444444444, 2)
        self.assertEqual(result, 2.94)

        result = colony.rounds(2.9944444444444444444444, 2)
        self.assertEqual(result, 2.99)

        result = colony.rounds(2.995, 2)
        self.assertEqual(result, 3.0)

        result = colony.rounds(1171.735, 2)
        self.assertEqual(result, 1171.74)

        result = colony.rounds(99999999.995, 2)
        self.assertEqual(result, 100000000.0)

        result = colony.rounds(999999999999999999999999999.9944444444444444444444, 2)
        if is_new: self.assertEqual(result, 999999999999999999999999999.99)

        result = colony.rounds(999999999999999999999999999.995, 2)
        if is_new: self.assertEqual(result, 1000000000000000000000000000.0)

        result = colony.rounds(770.155, 2)
        self.assertEqual(result, 770.16)

    def test_roundt(self):
        result = colony.roundt(2.675, 2)
        self.assertEqual(type(result), float)

        result = colony.roundt(colony.Decimal(2.675), 2)
        self.assertEqual(type(result), colony.Decimal)

        result = colony.roundt(2, 2)
        self.assertEqual(type(result), int)

    def test_apply(self):
        colony.round_apply(force = True)
        try:
            self.assertEqual(round, colony.roundi)
            result = round(2.675, 2)
            self.assertEqual(result, 2.68)
        finally:
            colony.round_unapply()

    def test_unapply(self):
        _round = round
        colony.round_apply(force = True)
        try: self.assertEqual(round, colony.roundi)
        finally: colony.round_unapply()
        self.assertEqual(round, _round)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Colony Framework
# Copyright (c) 2008-2020 Hive Solutions Lda.
#
# This file is part of Hive Colony Framework
#
# Hive Colony Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Colony Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Colony Framework If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2020 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import colony

class NumberTest(colony.ColonyTestCase):
    """
    Class that tests the number various functions method.
    """

    def test_to_fixed(self):
        """
        Tests the to fixed function.
        """

        # creates the the same value using an "infinite"
        # (repeating decimal) approach (through the 0.33)
        # and using the final and fixed value, this will create
        # problems in a normal float comparison
        infinite_float_value = 0.33 + 0.11 - 0.09 - 0.33
        correct_float_value = 0.02

        # verifies that the "infinite" (repeating decimal) based
        # float number is not the same as the non "infinite" based
        # number in a normal based float comparison
        self.assertNotEqual(infinite_float_value, correct_float_value)

        # converts both values into the fixed representation to test them
        # into a fixed based comparison, that must be valid
        infinite_fixed_value = colony.to_fixed(infinite_float_value, 2)
        correct_fixed_value = colony.to_fixed(correct_float_value, 2)

        # verifies that the comparison of the fixed based values should
        # be valid (this time the comparison takes no side effects)
        self.assertEqual(infinite_fixed_value, correct_fixed_value)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Colony Framework
# Copyright (c) 2008-2020 Hive Solutions Lda.
#
# This file is part of Hive Colony Framework
#
# Hive Colony Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Colony Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Colony Framework If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2020 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

from colony.base import legacy

from . import string_buffer_util

QUOTE_SAFE_CHAR = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789_.-"
""" The string containing all the safe characters to be quoted """

QUOTE_SAFE_MAPS = {}
""" The map of cached (buffered) safe lists to be quoted """

HEX_TO_CHAR_MAP = dict((legacy.bytes("%02x" % i), legacy.bytes(chr(i))) for i in range(256))
""" The map associating the hexadecimal byte (256) values
with the integers, the association is done using byte values """

# updates the map with the upper case values
HEX_TO_CHAR_MAP.update((legacy.bytes("%02X" % i), legacy.bytes(chr(i))) for i in range(256))

def quote(string_value, safe = "/"):
    """
    Quotes the given string value according to
    the URL encoding specification.
    The implementation is based on the python base library.

    :type string_value: String
    :param string_value: The string value to be quoted.
    :rtype: String
    :return: The quoted string value.
    """

    # in case the provided string value is unicode based it
    # must be encoded first using the default encoder
    is_unicode = type(string_value) == legacy.UNICODE
    if is_unicode: string_value = string_value.encode("utf-8")

    # creates the cache key tuple, that is going to be used
    # to avoid the re-creation of the safe map in every operation
    cache_key = (safe, QUOTE_SAFE_CHAR)

    try:
        # in case the cache key is not defined
        # in the quote safe maps, creates a new entry
        safe_map = QUOTE_SAFE_MAPS[cache_key]
    except KeyError:
        # adds the "base" quote safe characters to the
        # "safe list"
        safe += QUOTE_SAFE_CHAR

        # starts the safe map
        safe_map = {}

        # iterates over all the ascii values
        for index in range(256):
            # retrieves the character for the given index,
            # note that this strategy takes into account the
            # current version of the python environment
            character = chr(index)
            reference = index if legacy.PYTHON_3 else character

            # adds the "valid" character or the safe map entry
            safe_map[reference] = character if (character in safe) else ("%%%02X" % index)

        # sets the safe map in the cache quote safe maps
        QUOTE_SAFE_MAPS[cache_key] = safe_map

    # maps the get item method of the map to all the string
    # values to retrieve the valid items
    resolution_list = map(safe_map.__getitem__, string_value)

    # joins the resolution list to retrieve the quoted value
    # this will trigger the lazy loaded map operation
    return "".join(resolution_list)

def quote_plus(string_value, safe = ""):
    """
    Quotes the given string value according to
    the URL encoding specification. This kind of quote
    takes into account the plus and the space relation.
    The implementation is based on the python base library.

    :type string_value: String
    :param string_value: The string value to be quoted.
    :type safe: String
    :param safe: The string containing the characters considered
    safe for quoting.
    :rtype: String
    :return: The quoted string value.
    """

    # in case there is at least one white
    # space in the string value
    if " " in string_value:
        # quotes the string value adding the white space
        # to the "safe list"
        string_value = quote(string_value, safe + " ")

        # replaces the white spaces with plus signs and
        # returns the result
        return string_value.replace(" ", "+")

    # returns the quoted string value
    return quote(string_value, safe)

def unquote(string_value, strict = True):
    """
    Unquotes the given string value according to the URL
    encoding specification.
    The implementation is based on the python base library.

    :type string_value: String
    :param string_value: The string value to be unquoted, this
    string should be either encoded as an utf-8 based string
    or decoded with the proper unicode representation.
    :type strict: bool
    :param strict: If the unquoting operation should be performed
    using a strict approach meaning that all the characters are
    properly validated for ascii compliance, avoid duplicated
    unquoting operations.
    :rtype: String
    :return: The unquoted string value, this value is either
    returned as an utf-8 encoded string or an unicode string.
    """

    # forces the encoding of the string value as a string and
    # then splits the string value around percentage value
    # so that the various partial encoded values are decoded
    is_unicode = type(string_value) == legacy.UNICODE
    encoding = "ascii" if strict else "utf-8"
    if is_unicode: string_value = string_value.encode(encoding)
    if strict: string_value.decode("ascii")
    string_value_splitted = string_value.split(b"%")

    # iterates over all the "percentage values" range to decode
    # the complete set of percent encoded characters
    for index in range(1, len(string_value_splitted)):
        # retrieves the current iteration item, that is going to
        # be decoded using the hexadecimal to character map
        item = string_value_splitted[index]

        try:
            # tries to run the decoding operation using the hexadecimal
            # to character map and appends the remaining string value
            string_value_splitted[index] = HEX_TO_CHAR_MAP[item[:2]] + item[2:]
        except KeyError:
            # in case the decoding failed (no processing possible) the raw
            # value is set instead of the decoded value (fallback strategy)
            string_value_splitted[index] = b"%" + item

    # joins the various partial string values to be able to retrieve
    # the full unquoted utf-8 encoded string/bytes values, the value
    # is then decoded in case the current environment requires it
    unquoted = b"".join(string_value_splitted)
    return unquoted.decode("utf-8") if legacy.PYTHON_3 else unquoted

def unquote_plus(string_value, strict = True):
    """
    Unquotes the given string value according to
    the URL encoding specification. This kind of unquote
    takes into account the plus and the space relation.
    The implementation is based on the python base library.

    :type string_value: String
    :param string_value: The string value to be unquoted, this
    string should be either encoded as an utf-8 based string
    or decoded with the proper unicode representation.
    :type strict: bool
    :param strict: If the unquoting operation should be performed
    using a strict approach meaning that all the characters are
    properly validated for ascii compliance, avoid duplicated
    unquoting operations.
    :rtype: String
    :return: The unquoted string value.
    """

    # replaces the plus sign with a space, this is considered
    # the default plus symbol substitution, note that the proper
    # replacement operation is used according to base data type
    is_bytes = legacy.is_bytes(string_value)
    if is_bytes: string_value = string_value.replace(b"+", b" ")
    else: string_value = string_value.replace("+", " ")

    # returns the unquoted string value
    return unquote(string_value, strict = strict)

def url_encode(attributes_map = None, attributes_list = None, plus_encoding = False):
    """
    Encodes the given attributes into URL encoding. The
    attributes may be either provided as a map or alternatively
    as a sequence of key value tuples.

    :type attributes_map: Dictionary
    :param attributes_map: The map of attributes to be encoded
    using URL encoding, if this value is defined the attributes
    list is not going to be used.
    :type attributes_list: List
    :param attributes_list: The list of key and value tuples
    that are going to be processed as the attributes.
    :type plus_encoding: bool
    :param plus_encoding: If the plus encoding should be used.
    :rtype: String
    :return: The encoded attributes string.
    """

    # retrieves the quote method to be used
    quote_method = plus_encoding and quote_plus or quote

    # creates a string buffer to hold the encoded attribute values
    string_buffer = string_buffer_util.StringBuffer()

    # retrieves the reference to the proper items sequence to be used
    # notice that the attributes map has preference to the list version
    items = legacy.iteritems(attributes_map) if attributes_map else attributes_list

    # sets the is first flag, that is going to be used in the iteration
    # cycle to define the first iteration (exceptional iteration)
    is_first = True

    # iterates over all the attribute keys and values to be able to
    # quote their values and append the value to the buffer
    for attribute_key, attribute_value in items:
        # quotes both the attribute key and value
        attribute_key_quoted = quote_method(attribute_key)
        attribute_value_quoted = quote_method(attribute_value)

        # in case it's is the first iteration
        if is_first:
            # unsets the is first flag
            is_first = False
        else:
            # writes the and continuation in the string buffer
            string_buffer.write("&")

        # adds the quoted key and value strings to the
        # string buffer
        string_buffer.write(attribute_key_quoted)
        string_buffer.write("=")
        string_buffer.write(attribute_value_quoted)

    # retrieves the encoded attributes from the string buffer
    encoded_attributes = string_buffer.get_value()

    # returns the encoded attributes
    return encoded_attributes



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Colony Framework
# Copyright (c) 2008-2020 Hive Solutions Lda.
#
# This file is part of Hive Colony Framework
#
# Hive Colony Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Colony Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Colony Framework If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt> & Tiago Silva <tsilva@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2020 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import unittest

import xml.dom.minidom

from colony.base import legacy

class ColonyTestCase(unittest.TestCase):
    """
    The base class to be used for all the colony
    based test cases. It should contain a series
    of facilities aimed at simplifying the creation
    of unit tests under the colony infra-structure.
    """

    @staticmethod
    def get_test_case():
        return self

    @staticmethod
    def get_pre_conditions():
        return None

    @staticmethod
    def get_description():
        return "The base Colony Framework test case"

    def setUp(self):
        if not hasattr(self, "plugin"): return
        if not hasattr(self.plugin, "test"): return
        self.test = self.plugin.test
        self.test.set_up(self)

    def tearDown(self):
        if not hasattr(self, "plugin"): return
        if not hasattr(self.plugin, "test"): return
        self.test.tear_down(self)

    def assert_type(self, value, expected_type):
        """
        Tests that the provided value is of the specified type.

        :type value: Object
        :param value: The value whose type will be compared
        against the expected type.
        :type expected_type: Type
        :param expected_type: The type that is expected for
        the provided value.
        """

        # retrieves the value's type and in case the type is
        # the expected on returns immediately as no  problem
        # exists for the current assertion call
        value_type = type(value)
        if value_type == expected_type: return

        # raises a failure exception as the type of the value
        # is not the one expected by the assert operation
        raise self.failureException(
            "value is of type %s instead of expected type %s" %
            (value_type, expected_type)
        )

    def assert_raises(self, expected_exception, function, *args, **kwargs):
        """
        Tests that the specified exception is raised when invoking
        the provided function with the respective arguments.

        The expected exception parameter may contain the type of the
        expected exception or the name of it.

        :type expected_exception: Exception/String
        :param expected_exception: The type or name of the exception
        that should be raised by the function.
        :type function: Function
        :param function: The function to be invoked.
        """

        try:
            # invokes the function that should trigger the evaluation process,
            # note that the execution is dynamic with the provided arguments
            function(*args, **kwargs)
        except BaseException as exception:
            # checks if the current exception assert mode is string or value
            # oriented and then uses the string mode flag to find out the correct
            # expected exception name (for exception string description)
            string_mode = True if type(expected_exception) in legacy.STRINGS else False
            expected_exception_name = expected_exception if string_mode else expected_exception.__name__

            # retrieves the exception class and then uses it
            # to retrieve the exception name
            exception_class = exception.__class__
            exception_name = exception_class.__name__

            # re-raises the proper exception in case the raised exception was
            # not the expected one (so that the test may fail correctly) using
            # the string based comparison as it's currently set
            if string_mode and not exception_name == expected_exception_name: raise

            # re-raises the proper exception in case the raised exception was
            # not the expected one (so that the test may fail correctly) uses
            # the class type verification as no string mode is enabled
            if not string_mode and not exception_class == expected_exception: raise
        else:
            # raises a failure exception in case no exception was raised, this is
            # the expected behavior as an exception was expected in situation
            raise self.failureException("%s exception was not raised" % expected_exception)

    def assert_valid_xml(self, xml_data):
        """
        Tests that the provided XML data is valid.

        :type xml_data: String
        :param xml_data: The string with the XML data.
        """

        # attempts to parse the XML data
        # to check if its valid
        try:
            # parses the XML data
            xml.dom.minidom.parseString(xml_data)
        except Exception:
            # raises a failure exception
            # in case the parse was unsuccessful
            raise self.failureException("XML data is invalid")



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Colony Framework
# Copyright (c) 2008-2020 Hive Solutions Lda.
#
# This file is part of Hive Colony Framework
#
# Hive Colony Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Colony Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Colony Framework If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2020 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import copy

from colony.base import legacy

NOT_SET_VALUE = None
""" The value to be set when the value is not set (defined) """

TOPPER_VALUE = "_topper"
""" The value of the attribute to hold the top values """

LIST_TYPES = (list, tuple)
""" A tuple with the various list types """

INVALID_ATTRIBUTE_NAMES = ("__doc__", "__module__")
""" The set of invalid attribute names (for printing) """

VALID_ATTRIBUTE_TYPES = (
    int,
    legacy.LONG,
    float,
    bool,
    str,
    legacy.UNICODE,
    type(None)
)
""" The set of valid attribute types (for printing) """

def object_attribute_names(instance):
    """
    Function that retrieves the valid attribute
    names for printing.
    A valid attribute is considered to be an attribute
    that may have it's value ready for printing.

    :type instance: Object
    :param instance: The instance to retrieve the (valid)
    attribute names.
    :rtype: List
    :return: The valid attribute names according to the
    the state for printing the value.
    """

    # filters the attribute names based on the type and value
    # of them (non printable attributes are filtered out)
    valid_attribute_names = [key for key, value in __object_items(instance) if type(value) in VALID_ATTRIBUTE_TYPES]

    # returns the valid attribute names (ready for print)
    return valid_attribute_names

def object_attribute_values(instance, valid_attribute_names = None, strict = True):
    """
    Function that retrieves the valid attribute
    values for printing.

    A valid attribute is considered to be an attribute
    that may have it's value ready for printing.

    A non strict mode is possible to be used where in
    case an attribute is not found the value is set
    as not valid (invalid value).

    :type instance: Object
    :param instance: The instance to retrieve the (valid)
    attribute values.
    :type valid_attribute_names: List
    :param valid_attribute_names: List of valid attribute
    names that may be used to retrieve valid attribute values.
    :type strict: bool
    :param strict: If a strict based strategy should be used
    for the retrieval of the object attributes, meaning that
    an exception is raised in case the attribute does not exist.
    :rtype: List
    :return: The valid attribute value according to the
    the state for printing the value.
    """

    # retrieves the valid attribute names in order
    # to use them to retrieve the valid attribute values
    valid_attribute_names = valid_attribute_names or object_attribute_names(instance)
    valid_attribute_values = [__object_get_attr(instance, value, strict = strict) for value in valid_attribute_names]

    # returns the valid attribute values (ready for print)
    return valid_attribute_values

def object_flatten(instance, flattening_map):
    """
    Flattens the given instance using the given flattening
    map as reference for the flattening process.

    :type instance: Object
    :param instance: The instance to be flatten.
    :type flattening_map: Dictionary
    :param flattening_map: Map describing the structure
    for flattening.
    """

    # retrieves the type of the instance
    instance_type = type(instance)

    # in case the type of instance is (just) map (dictionary)
    # the instance should be encloses in a sequence object
    if instance_type == dict:
        # converts the instance to a list
        # (in order to be able to work with it)
        instance = [instance]
    # in case the instance is neither an instance
    # nor a list
    elif not instance_type in LIST_TYPES:
        # raises a runtime error (no valid instance type)
        raise RuntimeError("Invalid instance type")

    # flattens the structure of the instance (list)
    # using the flattening map (the returned structure
    # is a list of "flatten" instances)
    flatten_list = _object_flatten(instance, flattening_map)

    # returns the flatten list
    return flatten_list

def object_print_list(instances_list):
    """
    Prints some information on the objects
    in the given list, for debugging purposes.

    :type instances_list: List
    :param instances_list: The list of instances
    to be printed for debugging.
    """

    # iterates over all the instances
    # in the instances list to print them
    for instance in instances_list:
        # prints the debug information on
        # the instance
        object_print(instance)

        # prints a blank line
        print("")

def object_print(instance):
    """
    Prints some debug information on the
    instance, for debugging purposes.

    :type instance: Object
    :param instance: The instance to be printed
    for debugging.
    """

    # retrieves the list of all the attribute names
    # for the instance
    attribute_names = __object_keys(instance)

    # iterates over all the attribute names of the instance
    # (filters the invalid ones)
    for attribute_name in attribute_names:
        # retrieves the attribute and the name
        # of the attribute from the instance
        attribute = __object_get_attr(instance, attribute_name)
        attribute_type = type(attribute)

        # in case the attribute name or the type is invalid
        # must skip the current loop (not required)
        if attribute_name in INVALID_ATTRIBUTE_NAMES: continue
        if not attribute_type in VALID_ATTRIBUTE_TYPES: continue

        # prints the attribute name and the attribute value
        print("%s: %s" % (attribute_name, attribute))

def _object_flatten(instances_list, flattening_map):
    """
    Flattens the given instance using the given flattening
    map as reference for the flattening process.
    This function implements the concrete behavior for the
    flattening of an instance.

    :type instance: Object
    :param instance: The instance to be flatten.
    :type flattening_map: Dictionary
    :param flattening_map: Map describing the structure
    for flattening.
    :rtype: List
    :return: The list of instances in the flatten state.
    """

    # iterates over all the "base" instances
    for instance in instances_list:
        # flattens the instance in the to one relations
        # and the attributes (according to the flattening map)
        __object_flatten_to_one(instance, instance, flattening_map)

    # flattens the to many relation in the instances
    # in the given list
    instances_list = __object_flatten_to_many(instances_list, flattening_map)

    # flushes the "topper" map in the instances list
    __object_flush_topper(instances_list)

    # flushes the null elements in the instances list
    __object_flush_null(instances_list)

    # returns the list of flatten instances
    return instances_list

def __object_flatten_to_one(base_instance, instance, flattening_map):
    """
    Auxiliary function that provides the mechanism
    to "map" the "to-one" relations in the instance
    according to the flattening map.

    :type base_instance: Object
    :param base_instance: The base (top level) instance to
    be used to set the top level attributes.
    :type instance: Object
    :param instance: The current concrete instance in the
    recursion set.
    :type flattening_map: Dictionary
    :param flattening_map: Map describing the structure
    for flattening.
    """

    # iterates over all the keys and values
    # in the flattening map structure
    for key, value in legacy.iteritems(flattening_map):
        # retrieves the value type
        value_type = type(value)

        # tries to retrieves the instance value
        # sets the value to invalid in case there
        # is no such instance value (attribute)
        instance_value = __object_has_attr(instance, key) and __object_get_attr(instance, key) or None

        # in case the instance value is not set, must
        # continue the loop (skips the current iteration)
        if instance_value == None: continue

        # in case the value if of type string
        # (a leaf of the flattening structure)
        if value_type == str:
            # sets the leaf value in the base instance
            __object_set_attr(base_instance, value, instance_value)
        # in case the value is of type dictionary
        # (defined to one relation)
        elif value_type == dict:
            # "flattens" the to one instance relation (recursion)
            __object_flatten_to_one(base_instance, instance_value, value)

def __object_flatten_to_one_map(base_map, instance, flattening_map):
    """
    Auxiliary function that provides the mechanism
    to "map" the "to-one" relations in the instance
    according to the flattening map.

    :type base_map: Dictionary
    :param base_map: The base (top level) map to
    be used to set the top level attributes.
    :type instance: Object
    :param instance: The current concrete instance in the
    recursion set.
    :type flattening_map: Dictionary
    :param flattening_map: Map describing the structure
    for flattening.
    """

    # iterates over all the keys and values
    # in the flattening map structure
    for key, value in legacy.iteritems(flattening_map):
        # retrieves the value type
        value_type = type(value)

        # tries to retrieves the instance value
        # sets the value to invalid in case there
        # is no such instance value (attribute)
        instance_value = __object_has_attr(instance, key) and __object_get_attr(instance, key) or None

        # in case the instance value is not set
        if instance_value == None:
            # continues the loop
            continue

        # in case the value if of type string
        # (a leaf of the flattening structure)
        if value_type == str:
            # sets the leaf value in the base map
            base_map[value] = instance_value
        # in case the value is of type dictionary
        # (defined to one relation)
        elif value_type == dict:
            # "flattens" the to one instance relation (recursion)
            __object_flatten_to_one_map(base_map, instance_value, value)

def __object_flatten_to_many(instances_list, flattening_map):
    """
    Auxiliary function that provides the mechanism
    to "map" the "to-many" relations in the instance
    according to the flattening map.

    :type instances_list: List
    :param instances_list: The list of instances to have
    the "to-many" relations mapped.
    :type flattening_map: Dictionary
    :param flattening_map: Map describing the structure
    for flattening.
    :rtype: List
    :return: The resulting list after the "to-many" operations
    that include cartesian product.
    """

    # creates the new instances list
    new_instances_list = []

    # iterates over all the instance in the instances
    # list to process the to many relations
    for instance in instances_list:
        # creates a new (initial bucket)
        # with only the initial instance
        bucket = [instance]

        # retrieves all the attribute names (keys) for the instance
        attribute_names = __object_keys(instance)

        # retrieves all the to many attribute names of the instance based
        # on the type being a list type (tuple or list)
        to_many_attribute_names = [attribute_name for attribute_name in attribute_names if type(__object_get_attr(instance, attribute_name)) in LIST_TYPES]

        # iterates over all the "to many" attributes
        # to process the relations
        for to_many_attribute_name in to_many_attribute_names:
            # retrieves the to many attribute
            to_many_attribute = __object_get_attr(instance, to_many_attribute_name)

            # retrieves the (new) flattening map for the to many
            # attribute
            _flattening_map = flattening_map.get(to_many_attribute_name, {})

            # flattens the to many attribute (list) and retrieves the list
            # of to many instances list
            to_many_intances_list = __object_flatten_to_many(to_many_attribute, _flattening_map)

            # calculates the new bucket (list) based on the product
            # of the bucket against the to many instances list, this product
            # is made with the "help" of the new flattening map
            bucket = __object_flatten_product(bucket, to_many_intances_list, _flattening_map)

        # extends the new instances list with the bucket for
        # the current instance
        new_instances_list.extend(bucket)

    # returns the new instances list
    return new_instances_list

def __object_flush_topper(instances_list):
    """
    Flushes (clears) the temporary "topper"
    map in all the instances in the given list.

    :type instances_list: List
    :param instances_list: The list of instances to have
    the "topper" map cleared.
    """

    # iterates over all the instances in the
    # instances list (to clear the "topper" map)
    for instance in instances_list:
        # in case the instance does not
        # contain the topper map
        if not __object_has_attr(instance, TOPPER_VALUE):
            # continues th loop
            continue

        # retrieves the "topper" map for
        # the current instance
        _topper = __object_get_attr(instance, TOPPER_VALUE)

        # iterates over all the "topper" map
        # items (to set them in the instance)
        for key, value in legacy.iteritems(_topper):
            # sets the item in the instance
            __object_set_attr(instance, key, value)

        # deletes the (temporary) "topper"
        # map value
        __object_del_attr(instance, TOPPER_VALUE)

def __object_flush_null(instances_list):
    """
    Flushes (clears) the elements in the objects
    which don't have any value defined.
    This method allows every object to become uniform
    with the others.

    :type instances_list: List
    :param instances_list: The list of instances to have
    the null values flushed.
    """

    # creates the set of instance keys
    instances_keys = set()

    # iterates over all the instances in the
    # instances list (to clear retrieve the object keys)
    for instance in instances_list:
        # retrieves the keys (names) for the instance
        instance_keys = __object_keys(instance)

        # retrieves the instances keys from the
        # union of the instance keys element
        instances_keys = instances_keys.union(instance_keys)

    # iterates over all the instances in
    # the instance list
    for instance in instances_list:
        # iterates over all the instances keys
        # in the instances keys list
        for instances_key in instances_keys:
            # in case the instance already contains
            # the instances key
            if __object_has_attr(instance, instances_key):
                # continues the loop
                continue

            # sets the object key in the instance
            # as null (none)
            __object_set_attr(instance, instances_key, None)

def __object_flatten_product(first_list, second_list, flattening_map):
    """
    Provides a special case of the cartesian product of
    two sets (in this case lists).
    Both lists are "multiplied" and then using the flattening
    map the second list item is filtered accordingly.

    :type first_list: List
    :param first_list: The first (base) list for the cartesian
    product.
    :type second_list: List
    :param second_list: The second list for the cartesian product.
    :type flattening_map: Dictionary
    :param flattening_map: Map describing the structure
    for flattening.
    :see: http://en.wikipedia.org/wiki/Cartesian_product
    """

    # creates the initial product list
    product_list = []

    # iterates over all the items in
    # the first list
    for first_item in first_list:
        # iterates over all the items
        # in the second list
        for second_item in second_list:
            # creates a clone of an item
            # from the first (base) list
            new_item = copy.copy(first_item)

            # in case the second item contains
            # the "topper" attribute
            if __object_has_attr(second_item, TOPPER_VALUE):
                # retrieves the "topper" attribute
                # from the second item
                _topper = __object_get_attr(second_item, TOPPER_VALUE)
            # otherwise it's a leaf node and a "topper"
            # map must be created
            else:
                # creates a new "topper" map
                _topper = {}

            # flattens the to one relations in the second item
            # and puts them in the topper map
            __object_flatten_to_one_map(_topper, second_item, flattening_map)

            # sets the "topper" map in the new item
            __object_set_attr(new_item, TOPPER_VALUE, _topper)

            # adds the new item to the product
            # list
            product_list.append(new_item)

    # returns the (instance) product
    # list (result of multiplication)
    return product_list

def __object_has_attr(instance, attribute_name):
    """
    Checks if an attribute with the given name
    exists in the given instance.
    This method provides an additional layer of abstraction
    that allows it to be used in objects or in maps.

    :type instance: Object
    :param instance: The instance to be checked
    for attribute.
    :type attribute_name: String
    :param attribute_name: The name of the attribute
    to be checked in the instance.
    :rtype: bool
    :return: The result of the has attribute testing
    in the instance.
    """

    # retrieves the instance type
    instance_type = type(instance)

    # in case the instance type is dictionary
    if instance_type == dict:
        # checks if the attribute name exists in
        # instance (map)
        return attribute_name in instance
    # otherwise the instance is a "normal" instance
    else:
        # calls the normal has attr function
        # in the instance
        return hasattr(instance, attribute_name)

def __object_get_attr(instance, attribute_name, default = None, strict = True):
    """
    Retrieves an attribute with the given name from the
    given instance.

    This method provides an additional layer of abstraction
    that allows it to be used in objects or in maps.

    Note that the strict mode ensures that if the attribute
    is not present an exception is properly raised.

    :type instance: Object
    :param instance: The instance to retrieve the attribute, this
    may be either a dictionary or an object instance.
    :type attribute_name: String
    :param attribute_name: The name of the attribute to be retrieved
    from the instance.
    :type default: Object
    :param default: The default value to be returned in the value is
    not found, this is only used in case the strict mode is not set.
    :type strict: bool
    :param strict: If the strict mode should be used, meaning that
    the attributes must be set under the instance otherwise an exception
    will be raised.
    :rtype: Object
    :return: The retrieved attribute from the instance.
    """

    # retrieves the instance type
    instance_type = type(instance)

    # in case the instance type is dictionary the typical
    # dictionary access operation is used
    if instance_type == dict:
        if strict: return instance[attribute_name]
        else: return instance.get(attribute_name, default)

    # otherwise the instance is a "normal" instance, note
    # that the attribute is first verified for presence
    # in case the non strict mode is currently in use
    else:
        if strict: return getattr(instance, attribute_name)
        elif hasattr(instance, attribute_name):
            return getattr(instance, attribute_name)
        else: return default

def __object_set_attr(instance, attribute_name, attribute):
    """
    Sets an attribute with the given name in the
    given instance.
    This method provides an additional layer of abstraction
    that allows it to be used in objects or in maps.

    :type instance: Object
    :param instance: The instance to retrieve the
    attribute.
    :type attribute_name: String
    :param attribute_name: The name of the attribute
    to be set in the instance.
    :type attribute: Object
    :param attribute: The attribute (value) to be set in
    the instance.
    """

    # retrieves the instance type
    instance_type = type(instance)

    # in case the instance type is dictionary
    if instance_type == dict:
        # sets the attribute using the normal
        # map setter
        instance[attribute_name] = attribute
    # otherwise the instance is a "normal" instance
    else:
        # uses the typical setattr function to set
        # the attribute in the instance
        setattr(instance, attribute_name, attribute)

def __object_del_attr(instance, attribute_name):
    """
    Deletes the attribute with the given name from the
    given instance.
    This method provides an additional layer of abstraction
    that allows it to be used in objects or in maps.

    :type instance: Object
    :param instance: The instance to delete the
    attribute.
    :type attribute_name: String
    :param attribute_name: The name of the attribute
    to be deleted.
    :rtype: Object
    :return: The retrieved attribute from the instance.
    """

    # retrieves the instance type
    instance_type = type(instance)

    # in case the instance type is dictionary
    if instance_type == dict:
        # calls the del operator in the instance
        # map (dictionary)
        del instance[attribute_name]
    # otherwise the instance is a "normal" instance
    else:
        # deletes the attribute from the instance
        # using the delattr function
        delattr(instance, attribute_name)

def __object_keys(instance):
    """
    Retrieves a list with all the instance names (keys),
    from the given instance.
    This method provides an additional layer of abstraction
    that allows it to be used in objects or in maps.

    :type instance: Object
    :param instance: The instance to retrieve the keys
    list (names list).
    """

    # retrieves the instance type
    instance_type = type(instance)

    # in case the instance type is dictionary
    if instance_type == dict:
        # returns the instance (map) keys values
        # using the normal map method
        return legacy.keys(instance)
    # otherwise the instance is a "normal" instance
    else:
        # returns the instance dictionary keys
        # (the instance names)
        return legacy.keys(instance)

def __object_items(instance):
    """
    Retrieves a list with all the instance value and keys (items),
    from the given instance.
    This method provides an additional layer of abstraction
    that allows it to be used in objects or in maps.

    :type instance: Object
    :param instance: The instance to retrieve the values
    and keys list (items list).
    """

    # retrieves the instance type
    instance_type = type(instance)

    # in case the instance type is dictionary
    if instance_type == dict:
        # returns the instance (map) keys values
        # using the normal map method
        return legacy.items(instance)
    # otherwise the instance is a "normal" instance
    else:
        # returns the instance dictionary keys
        # (the instance names)
        return legacy.items(instance.__dict__)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Colony Framework
# Copyright (c) 2008-2020 Hive Solutions Lda.
#
# This file is part of Hive Colony Framework
#
# Hive Colony Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Colony Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Colony Framework If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2020 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import binascii

from colony.base import legacy

def encode_two_complement_string(long_value):
    """
    Encode a long to a two's complement little-endian binary string.
    Note that "0L" is a special case, returning an empty string, to save a
    byte.

    :type long_value: int
    :param long_value: The long value to be encoded.
    :rtype: String
    :return: The encoded two's complement little-endian binary string.
    """

    # in case the long value is zero
    if long_value == 0:
        # returns empty string
        return ""
    # in case the long value is larger
    # than zero
    elif long_value > 0:
        # converts the long value to the hexadecimal string value
        long_value_hexadecial = hex(long_value)

        # counts the number of nibbles in the given value
        number_nibbles = _count_nibbles(long_value_hexadecial)

        # in case the number of nibbles is odd
        if number_nibbles & 1:
            # need an even # of nibbles for unhexlify
            long_value_hexadecial = "0x0" + long_value_hexadecial[2:]
        # in case the number i
        elif int(long_value_hexadecial[2], 16) >= 8:
            # looks negative so need a byte of sign bits
            long_value_hexadecial = "0x00" + long_value_hexadecial[2:]
    else:
        # converts the (negative) long value to the hexadecimal string value
        long_value_hexadecial = hex(-long_value)

        # counts the number of nibbles in the given value
        number_nibbles = _count_nibbles(long_value_hexadecial)

        # in case the number of nibbles is odd
        if number_nibbles & 1:
            # extends to a full byte.
            number_nibbles += 1

        # calculates the number of bits from the nibbles
        number_bits = number_nibbles * 4

        # puts the negative indication as the last digit
        long_value += 1 << number_bits

        # converts the long value to the hexadecimal string value
        long_value_hexadecial = hex(long_value)

        # counts the number of nibbles in the given value
        new_number_nibbles = _count_nibbles(long_value_hexadecial)

        # in case the new number of nibbles is smaller than
        # the previous one
        if new_number_nibbles < number_nibbles:
            # puts the sign bits
            long_value_hexadecial = "0x" + "0" * (number_nibbles - new_number_nibbles) + long_value_hexadecial[2:]
        # in case it's a positive number
        if int(long_value_hexadecial[2], 16) < 8:
            # no need a byte of sign bits
            long_value_hexadecial = "0xff" + long_value_hexadecial[2:]

    # in case the long value hexadecimal ends with the long
    # indication value
    if long_value_hexadecial.endswith("L"):
        # removes the last character to avoid the extra long indication
        long_value_hexadecial = long_value_hexadecial[2:-1]
    # otherwise
    else:
        # sets the normal value without the "0x" initialization
        long_value_hexadecial = long_value_hexadecial[2:]

    # unhexflies the value retrieving the binary value
    binary = binascii.unhexlify(long_value_hexadecial)

    # reverses the binary value
    reversed_binary = binary[::-1]

    # returns the reversed binary value
    return reversed_binary

def decode_two_complement_string(data):
    """
    Decode a long from a two's complement little-endian binary string.

    :type data: String
    :param data: The data to be used in the decoding.
    :rtype: int
    :return: The decoded data.
    """

    # retrieves the data length
    data_length = len(data)

    # in case the data length is zero
    if data_length == 0:
        # return zero
        return legacy.LONG(0)

    # converts the (inverted) data to hexadecimal string
    long_value_hexadecial = binascii.hexlify(data[::-1])

    # converts the long value hexadecimal to integer
    # using base 16
    long_value = legacy.LONG(long_value_hexadecial, 16)

    # in case the last digit is 0x80 (negative)
    if data[-1] >= "\x80":
        # puts the negative indication as the last digit
        long_value -= legacy.LONG(1) << (data_length * 8)

    # returns the long value
    return long_value

def _count_nibbles(long_value_hexadecial):
    """
    Calculates the number of nibbles (4 bit group) from the given
    long value encoded in hexadecimal string.

    :type long_value_hexadecial: String
    :param long_value_hexadecial: The long value encoded in
    hexadecimal string to count the nibbles.
    :rtype: int
    :return: The number of nibbles in the given value.
    """

    # calculates the number of "junk" characters
    number_junk_characters = 2 + long_value_hexadecial.endswith("L")

    # calculates the number of nibbles from the length of the
    # long value string without the junk characters
    number_nibbles = len(long_value_hexadecial) - number_junk_characters

    # returns the number of nibbles
    return number_nibbles
